/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Log: parse_tag.c,v $
 * Revision 1.23  2000/04/18 15:11:36  gellert
 * unallowed_href returned -1 if the attribute had no value, so
 * attributes without values would not be allowed. fixed.
 *
 * Revision 1.22  2000/04/16 15:38:58  gellert
 * fixed two bugs:
 * a) check_for_script was not called. and now it checks all the
 *    ampersands for a following brace, not just the first one.
 * b) we have to call proto_allowed not only for HREF, but also
 *    for the SRC attribute.
 *
 * Revision 1.21  2000/03/07 15:06:54  goldbach
 * added GPL info
 *
 * Revision 1.20  2000/02/21 14:11:52  gellert
 * And we shouldn't call logmsg with NULL-pointers... ;-)
 *
 * Revision 1.19  2000/02/21 13:57:50  gellert
 * In "unallowed_href": inserted a test, if the given value is
 * a NULL-pointer. Otherwise httpf segfaults at <A HREF>...
 *
 * Revision 1.18  2000/02/21 13:42:47  gellert
 * In the valeu of a HREF we now search not only for ":", but
 * also for &#58, to prevent the values "javascript&#58"
 *
 * Revision 1.17  1999/11/26 10:47:08  goldbach
 * New actions introduced: replskip and replendskip skip sections of documents, i.e. data won\'t be sent to the client.
 *
 * Revision 1.16  1999/11/12 12:16:20  gellert
 * Fixed:	If an attribute was logged only (and not replaced or
 * 	allowed), the attributes were processed afterwards
 * 	(so the "<" and the tag were ommitted, but the attributes
 * 	and the ">" were copied to the outbuf).
 *
 * Revision 1.15  1999/11/11 15:58:02  gellert
 * Removed declarations of some nonexistent functions.
 * Some new logmessages inserted.
 *
 * Revision 1.14  1999/08/09 12:09:32  goldbach
 * Changed unallowed_href to accept hrefs without *any* protocols at all :)
 * Bug: do_attr aborted after processing an attribute without a value. This
 * buggy behaviour must have been around since the removal of the high-level
 * strncpy() for everything between the angles... tables look fine now :)
 *
 * Revision 1.13  1999/07/28 14:43:05  gellert
 * moved some functions to misc.c
 *
 * Revision 1.12  1999/07/22 15:35:04  goldbach
 * yet again cleanup for logmsg()
 *
 * Revision 1.11  1999/07/22 14:30:52  gellert
 * some improvements of the determination of the tag end,
 * now nested quotes etc are supported.
 * changed config option replace to replabort/replcont.
 *
 * Revision 1.10  1999/07/15 17:01:22  gellert
 * added the check of href tags for allowed protocols.
 *
 * Revision 1.9  1999/07/14 09:50:06  goldbach
 * removed last strncpy.
 *
 * Revision 1.8  1999/06/30 14:10:15  goldbach
 * removed subtle bug in process attr
 *
 * Revision 1.7  1999/06/29 11:56:41  gellert
 * parsing changed: now there are 0-bytes written into the
 * input buffer to terminate the strings. so they need not
 * to be copied into some extra buffer...
 *
 * Revision 1.6  1999/06/23 11:14:40  gellert
 * deleted some never reached line
 *
 * Revision 1.5  1999/06/23 10:45:20  gellert
 * better checking of the tag,
 * bug fixes (segfaults).
 *
 *
 * $Id: parse_tag.c,v 1.23 2000/04/18 15:11:36 gellert Exp $
 */

#include <syslog.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <ctype.h>
#include "mythreads.h"
#include "taghash.h"
#include "misc.h"

int parse_tag(char *, TCT *);

char *do_attr(char *, TCT *);
int process_tag(char *, TCT *);
char *find_tagstart(char *);
char *process_attr(char*, char *, TCT *);
int check_tag(char *, TCT*);

int unallowed_href(char *, char *);

/*
 * unallowed_href
 *
 * Checks if a hyperlink is allowed or not. It is allowed if it's
 * a reference without protocol or a reference with a
 * configured protocol.
 * Note:
 *   Currently this works only with the colon character ':' as protocol
 *   delimiter. We search &#xx, too. At least &#58 (that's ":")
 *   If the protocol name contains any &#xx, proto_allowed will not
 *   find them (and thus not allow them)
 */
int unallowed_href(char *attr, char *value) {
    char *temp, *temp1, *temp2;

    /* do we have a value? */
    if (value == NULL) {
      return(0);
    }

    /* do we have a link or a src? */
    if ((strncasecmp("href", attr, 5) != 0) &&
        (strncasecmp("src", attr, 4) != 0)) {
	return(0);
    }
    
    /* look for a colon in the attribute's value */
    temp1 = strnchr(value, ':', strlen(value));
    temp2 = strstr(value, "&#58");

    if ((temp1 == NULL) && (temp2 == NULL)) {
	/* there is no colon */
	/* we don't have a protocol, so the reference is okay */
	return(0);
    }

    temp = value;
    while ((*temp== ' ') || (*temp== '\"') || (*temp=='\'')) {
	temp++;
    }
    
    
    if (proto_allowed(temp)) {
	return(-1);
    } else {
	return(0);
    }
}



/*
 * find_tagstart
 *
 * Find start of tag.
 * See find_attrstart, the only difference is that we allow
 * a '/' at the start. 
 */
char *find_tagstart(char *buf) 
{
    char *tag_start;  

    /* skip whitespace */
    tag_start = skip_leading_whitespace(buf);
    
    if(legal_constituent(tag_start[0]) || tag_start[0]=='/')
	return tag_start;
    else
	return NULL;
}

/*
 * find_attrstart
 *
 * Find start of attribute.
 * Takes pointer to buffer that hopefully contains attributes.
 * Leading whitespace is skipped.
 * Returns pointer to first character of attribute or NULL if
 * attribute starts with illegal characters.
 */
char *find_attrstart(char *buf) 
{
    char *attr_start;  

    /* skip whitespace */
    attr_start = skip_leading_whitespace(buf);
    
    if(legal_constituent(attr_start[0]))
	return attr_start;
    else
	return NULL;
}

/*
 * find_attrend
 *
 * Find end of attribute.
 * Takes pointer to beginning of attribute. 
 * Return pointer to last character of attribute or NULL if
 * attribute contains illegal characters.
 */
char *find_attrend(char *attr_start) 
{
    int i=0;
    
    /*
     * Find end of legal characters.
     * This should be the attribute name.
     */

    while(legal_constituent(attr_start[i])) {
	i+=1;
    }

    /*
     * an attribute name ends with
     *  - null character
     *      this is last attribute for this tag
     *  - '='
     *      we have a value
     *  - whitespace
     *      we may have another attr or a value.
     *      the HTML specs allow only this: "attr=value",
     *      but Alas!, the world does "attr =  value".
     */
    if (attr_start[i]!='=' &&
	attr_start[i]!=0 &&
	legal_whitespace(attr_start[i])==NULL)
	return NULL;
    else
	return attr_start+i;
}

/*
 * find_valstart
 *
 * Find start of (optional) value for an attribute.
 * BUF points to the start of the current attribute.
 *
 * Returns:
 *  0 - no value
 *  other - start of value.
 */

char *find_valstart(char *attr_end)
{
    char *cont;

    /* we should have a '=' or a whitespace now */
    if (legal_whitespace(attr_end[0])!=NULL) {
	cont=skip_leading_whitespace(attr_end);
    } else
	cont=attr_end;

    /* no value if we are at end of string */
    if (cont[0]==0) return NULL;
    
    /*
     * now the only valid continuations are:
     *   '=' -- we have a value
     *   legal_constituent -- new attribute
     * Other values are an error.
     */
    if (cont[0]=='=') {
	/* we have a value */
	cont = skip_leading_whitespace(cont+1);
	if (legal_constituent(cont[0])!=NULL ||
            a_quote(cont))
	    /* valstart is legal */
	    return cont;
	else
	    /* valstart is illegal */
	    return (char *)-1;
    }

    /* we have no value or a malformed input */
    if (legal_constituent(cont[0])!=NULL)
	return NULL; /* no value */
    else
	return (char *)-1; /* illegal continuation */
}

/*
 * do_attr
 *
 * Try to get and handle the next attribute in a string.
 * Returns:
 *  0  -- abort parsing this string. 
 *        The attr is not allowed so skip the rest.
 *  -1 -- there was an error (illegal characters in attribute or value).
 *  other --  the string right behind the attribute's value or the string
 *            right behind the attribute (if there was no value)
 *
 * Note: we modify the input buffer. The attribute and value string
 *       are terminated with a null character.
 */
char *do_attr(char *buf, TCT *tct) 
{
    char *attr_start, *attr_end, *val_start, *val_end;
    char replaced;

    /* find attribute */
    attr_start = find_attrstart(buf);
    if (attr_start==NULL)  /* illegal chars? */
	return (char *)-1;
    
    attr_end = find_attrend(attr_start);
    if (attr_end==NULL)  /* illegal chars? */
	return (char *)-1;
        
    /* find value */
    val_start=find_valstart(attr_end);
    if (val_start==NULL) {
	/* no value */
	replaced=attr_end[0]; /* store char */
	attr_end[0]=0; /* terminate attribute string */
        logmsg(LOG_NOTICE, "[%d] Processing attribute: %s", tct->tid, attr_start);
	process_attr(attr_start, NULL, tct);	
	attr_end[0]=replaced; /* put it back */
	return attr_end;
    } else if (val_start==(char *)-1) {
	/* illegal value */
	logmsg(LOG_NOTICE, "[%d] Value for attribute '%s' contains illegal chars.",
	       tct->tid, attr_start);
	return (char *)-1;
    }

    val_end=find_valend(val_start, tct);
    if (val_end==NULL) {
	/* illegal value */
	logmsg(LOG_NOTICE, "[%d] Value for attribute '%s' contains illegal chars.",
	       tct->tid, attr_start);
	return (char *)-1;
    }
    
    replaced=val_end[0]; /* remember the char we're about to replace */
    attr_end[0]=0; val_end[0]=0;    /* terminate attribute and value */
    logmsg(LOG_NOTICE, "[%d] Processing attribute %s with value %s", tct->tid, attr_start, val_start);
    process_attr(attr_start, val_start,tct);     /* process it */
    /* put it back */
    val_end[0]=replaced;
    
    return val_end;
}

/*
 * do_tag
 *
 * Try to get and handle a tag in string. This is basically the same as
 * do_attr() except that a tag is slightly more than an attribute
 * (characterwise) -- the slash ('/') is a legal character, too.
 * See do_attr() for a description and notes.
 */
char *do_tag(char *buf, TCT *tct) 
{
    char *tag_start, *tag_end;
    int tag_result; char replaced;
    
    /* find tag */
    tag_start = find_tagstart(buf);
    if (tag_start==NULL) {
	logmsg(LOG_NOTICE, "[%d] Tag '%s' contains illegal characters.",
	       tct->tid, buf);
	return (char *)-1;
    }
    
    /*
     * find end of tag. this is the same as finding the end
     * of an attribute so we use the same function.
     */
    tag_end = find_attrend(tag_start+1);
    /* illegal ? */
    if (tag_end==NULL) {
	logmsg(LOG_NOTICE, "[%d] Tag '%s' contains illegal characters.",
	       tct->tid, buf);
	return (char *)-1;
    }
    
    /* terminate tag. remember the char we replace */
    replaced=tag_end[0];
    tag_end[0]=0;
    tag_result=process_tag(tag_start, tct);
    /* put back what we replaced. this is important for the caller. */
    tag_end[0]=replaced;

    if (tag_result==-1)
	/* abort: tag not allowed */
 	return (char *)-1;
    else
	/* okay, continue with attributes */
	return tag_end; /* return addr. of last char */
}

/*
 * check_tag
 *
 * Performs a very rough check of the "tag":
 *   - comments are replaced with a short standard note
 *   - !DOCTYPE... is copied
 *   - others are left untouched.
 */
int check_tag(char *buf, TCT *tct) 
{
    /* is it a comment? */
    if (strncmp(buf, "!--",3)==0) {
	/* yes, it is: skip it */
	output_string("<!-- comment suppressed -->", 27, tct);
	return -1;
    }
    
    /* is a doctype spec.? */
    if (strncmp(buf, "!DOCTYPE",8)==0) {
	/* yes, it is: just copy it to output buffer */
	output_string("<", 1, tct);
	output_string(buf, strlen(buf), tct);
	output_string(">", 1, tct);
	return -1;
    }

    /* otherwise expect to have a normal tag */
    return 0;
}


/*
 * parse_tag
 *
 * Parse the tag in the buffer.
 */
int parse_tag(char *buf, TCT *tct) 
{
    char *s=buf;

    /* check if we should parse or silently skip it */
    if (check_tag(buf, tct)==-1) return 0; 
    
    logmsg(LOG_DEBUG, "[%d] Start splitting attributes in '%s'.",
	   tct->tid, buf);
    
    s=do_tag(s,tct);
    logmsg(LOG_DEBUG, "[%d] Processed tag.", tct->tid);
    
    /* check for error */
    if (s==(char *)-1) {
	/* error */
	return -1; 
    } else if (s==NULL) {
	/* skip */
 	output_string(">", 1, tct); /* 'close' it */
    }
    
    /* check if there are any attrs to process */
    if (s[0]==0) {
	/* no */
	logmsg(LOG_DEBUG, "[%d] This tag has no attributes.", tct->tid);
 	output_string(">", 1, tct); /* 'close' the tag */
	return 0;
    } else {
	/* we know there are attributes */
	logmsg(LOG_DEBUG, "[%d] Processing attributes in '%s'.", tct->tid, s);
 	while(s!= NULL && s!=(char *)-1) { /* s>0 doesn't work */ 
	    s=do_attr(s,tct);
	}
	logmsg(LOG_DEBUG, "[%d] Processed attributes.", tct->tid);
	output_string(">", 1, tct);
    }
    
    return 0;
}


/*
 * process_tag
 *
 * Takes buffer containing the tag and the context.
 * Look up the tag in hash-table built from the configuration and
 * act like the action says.
 * Returns:
 *  0  -- continue with the attributes
 * -1  -- tag not allowed, skip rest of tag string
 */
int process_tag(char *tagbuf, TCT *tct) {
    char *replacement; /* from hash-table */
    int action; /* from hash-table */
    
    /* find out what we are supposed to do */
    action = tag_action(tagbuf);
    
    if (action < 0) {
	/* not specified in configuration file */
 	logmsg(LOG_NOTICE, "[%d] Not allowing tag '%s'.", tct->tid, tagbuf);
	return -1;
    }
    
    if (action & (ACT_REPLABORT | ACT_REPLCONT | ACT_REPLSKIP |
                  ACT_REPLENDSKIP)) {
	/* get replacement for this tag*/
	replacement=tag_replacement(tagbuf);
	if (action & ACT_LOG) {
	    logmsg(LOG_NOTICE, "[%d] LOG: Replacing %s with '%s'.",
		   tct->tid, tagbuf, replacement);
	}
	
	output_string("<", 1, tct);
	output_string(replacement, strlen(replacement),tct);
		
	if (action & ACT_REPLSKIP) {
	    output_string(">", 1, tct);
	    tct->skip_flag=1; /* skip the following contents */
	    return -1; /* don't proceed with the attributes */
	} else if (action & ACT_REPLENDSKIP) {
	    /* turn off skipping mode and write the replacement */
	    tct->skip_flag=0;
	    output_string("<", 1, tct);
	    output_string(replacement, strlen(replacement),tct);
	    output_string(">", 1, tct);
	    
	    return -1; /* don't proceed with the attributes */
	} else if (action & ACT_REPLABORT) {
	    output_string(">", 1, tct);
	    return -1; /* don't proceed with the attributes */
	} else {
	    return 0; /* proceed with the attributes */
	}
    } else if (action & ACT_ALLOW) {
	if (action & ACT_LOG) {
	    logmsg(LOG_NOTICE, "[%d] LOG: Allowing tag %s.", tct->tid, tagbuf);
	}
	
	/* copy tag 'as is' to output buffer */
	output_string("<", 1, tct);
	output_string(tagbuf, strlen(tagbuf), tct);
	return 0; /* proceed with the attributes */
    } else if (action & ACT_LOG) {
	/* just log it */
	logmsg(LOG_NOTICE, "[%d] LOG: Skipping tag: %s", tct->tid, tagbuf);
	return -1; /* don't proceed with the attributes */
    }
    
    logmsg(LOG_ERR, "[%d] Action %d not processed", tct->tid, action);
    logmsg(LOG_ERR, "[%d] This is a bug: FIXME", tct->tid);
    return -1;
}


/*
 * process_attr
 *
 */
char *process_attr(char*attrbuf, char *valbuf, TCT *tct) 
{
    char *replacement; /* from hash-table */
    int action; /* from hash-table */
    
    /* find out what we are supposed to do */
    action = attrib_action(attrbuf);
    
    if (action < 0) {
	/* action not specified in configuration file */
 	logmsg(LOG_NOTICE, "[%d] Not allowing attribute '%s'.", tct->tid, attrbuf);
	return (char *)-1;
    }

    /* check if we have some script code in the value */
    if (check_for_script(valbuf)!=0){
	logmsg(LOG_NOTICE, "Found script code in value of attribute %s:", attrbuf);
	logmsg(LOG_NOTICE, "\"%s\"", valbuf);
	return (char *)-1;
    }

    /* logging? */
    {
	int level;
	if (action & ACT_LOG)
	    level=LOG_NOTICE;
	else
	    level=LOG_DEBUG;
	
	if (valbuf!=NULL)
	    logmsg(level, "[%d] Processing attr '%s' with value '%s'.",
		   tct->tid, attrbuf, valbuf);
	else
	    logmsg(level, "[%d] Processing attr '%s'.",
		   tct->tid, attrbuf);
    }


    /* you can either replace an attr or allow it */

    if (action & (ACT_REPLABORT | ACT_REPLCONT)) {
	  /* get replacement for this attr*/
	  replacement=attrib_replacement(attrbuf);
	  if (action & ACT_LOG) {
	      logmsg(LOG_NOTICE, "[%d] Replacing with '%s'.",
		     tct->tid, replacement);
	  }
	  
	  output_string(" ", 1, tct);
	  output_string(replacement, strlen(replacement),tct);
	  if (action & ACT_REPLABORT) {
	    return (char *)-1;
	  } else {
	    return 0; /* proceed with the attributes */
	  }
    } else if (action & ACT_ALLOW) {
       if (unallowed_href(attrbuf, valbuf)) {
          if (valbuf == NULL) {
              logmsg(LOG_NOTICE, "[%d] Unallowed HREF: no attribute", tct->tid);
	  } else {
              logmsg(LOG_NOTICE, "[%d] Unallowed HREF: '%s'", tct->tid, valbuf);
          }
       } else {
	  if (action & ACT_LOG) {
	      logmsg(LOG_NOTICE, "[%d] Allowing attribute.", tct->tid);
	  }
	  
	  /* copy attr 'as is' to output buffer */
	  output_string(" ", 1, tct);
	  output_string(attrbuf, strlen(attrbuf), tct);
	  if (valbuf!=NULL) {
	      /* we have a value */
	      output_string("=", 1, tct);
	      output_string(valbuf, strlen(valbuf), tct);
	  }
       }
       return 0; /* proceed with the attributes */
    } else if (action & ACT_LOG) {
	/* just log it */
	/*logmsg(LOG_NOTICE, " (no further action)"); */
	return 0; /* proceed with the attributes */
    }
    
    logmsg(LOG_ERR, "[%d] Action %d not processed.", tct->tid, action);
    logmsg(LOG_ERR, "[%d] This is a bug: FIXME.", tct->tid);
    return (char *)-1;
    
	    
}

/*
 * check_for_script
 *
 * Checks the value buffer for embedded script code.
 * (This script code is introduced by "&{".)
 * Returns 0 if the buffer does not contain "&{",
 * -1 otherwise.
 * Notes: Safe to call with a value of NULL for the buffer.
 */
int check_for_script(char *valbuf) 
{
    char *pos;

    /* return ok if we have no value */
    if (valbuf==NULL)
	return 0;

    pos = valbuf;

    /* look for ampersand */
    while ((pos = strnchr(pos, '&', strlen(pos))) != NULL) {
      if (strlen(pos)>1) {
	if (pos[1]=='{') return -1;	
      }
      pos++;
    }

    /* no script code found */
    return 0;
}
